module ClassConstant : sig
	type t
	val idx : t -> int
	val name : t -> string
	val to_string : t -> string
	val create : int -> t
	val update_ref : string -> t -> t
	val empty : t
end;;

module NATConstant : sig
	type t
	val name_idx : t -> int
	val desc_idx : t -> int
	val name : t -> string
	val desc : t -> string
	val to_string : t -> string
	val create : int -> int -> t
	val update_ref : string -> string -> t -> t
	val empty : t
end;;

module FMIConstant : sig
	type refType 
	type t
  val fmi_type : t -> refType
  val class_idx : t -> int
  val nat_idx : t -> int
  val classConstant : t -> ClassConstant.t
  val natConstant : t -> NATConstant.t
  val to_string : t -> string
  val create : refType -> int -> int -> t
	val update_ref : ClassConstant.t -> NATConstant.t -> t -> t
end;;

module StringConstant : sig
	type t
  val idx : t -> int
  val str : t -> string
  val to_string : t -> string
  val create : int -> t
	val update_ref : string -> t -> t
end;;

module ValueConstant : sig 
	type t
  val to_string : t -> string
  val create : int -> t
	val value : t -> int
end;;

module UTF8Constant : sig
	type t
  val length : t -> int
  val bytes : t -> int list
  val utf_string : t -> string
  val to_string : t -> string
  val create : int -> int list -> string -> t
end;;

type t =
		Class of ClassConstant.t
	| NAT of NATConstant.t
	| FMI of FMIConstant.t
	| String of StringConstant.t
	| Value of ValueConstant.t
	| UTF8 of UTF8Constant.t;;
val to_string : t -> string;;

val newConstantClass : int -> t
val newConstantNat : int -> int -> t
val newConstantFieldRef : int -> int -> t 
val newConstantMethodRef : int -> int -> t
val newConstantInterfaceMethodRef : int -> int -> t
val newConstantString : int -> t
val newConstantValue : int -> t
val newConstantUtf : int -> int list -> string -> t

val getConstantUtf : int -> (int -> t) -> UTF8Constant.t;;
val getConstantClass : int -> (int -> t) -> ClassConstant.t;;
val getConstantValue : int -> (int -> t) -> ValueConstant.t;;

(*
type constantType = Class | FieldRef | MethodRef | InterfaceMethodRef | String | Integer | NameAndType | Utf8 | MethodHandle | MethodType
(* | Float | Long | Double | InvokeDynamic  Unused *) ;;

type class_t = { className_index: int; actual_className: string };;
type nat_t = { natName_index: int; descriptor_index: int;
	actual_natName: string; actual_descriptor: string };;
(* fmi = shorthand for field, method, and interfaceMethod *)
type fmiRef_t = { class_index: int; nat_index: int; actual_class: class_t; actual_nat: nat_t };;
type string_t = { string_index: int; actual_string: string };;
type value_t = { value: int };;
(* wideType = {high_bytes:int; low_bytes:int};; *)
type utf_t = { length: int; bytes: int list; utf_string: string };;
(* type methodHandleType = {};; type methodTypeType = {};; type            *)
(* invokeDynamic = {};;                                                    *)

type 'a constant_t = { tag: constantType; info:'a };;

type constant =
		ConstantClass of class_t constant_t
	| ConstantNat of nat_t constant_t
	| ConstantFmi of fmiRef_t constant_t
	| ConstantString of string_t constant_t
	| ConstantValue of value_t constant_t
	| ConstantUtf of utf_t constant_t;;

(* uninitialized values *)
val newClassInfo : class_t
val newNatInfo : nat_t
val newFmiRefInfo : fmiRef_t
val newStringInfo : string_t
val newValueInfo : value_t
val newUtfInfo : utf_t

val newConstantClass : class_t constant_t
val newConstantNat : nat_t constant_t
val newConstantFieldRef : fmiRef_t constant_t
val newConstantMethodRef : fmiRef_t constant_t
val newConstantInterfaceMethodRef : fmiRef_t constant_t
val newConstantString : string_t constant_t
val newConstantValue : value_t constant_t
val newConstantUtf : utf_t constant_t

(* String Converter *)
val string_of_constant : constant -> string
val string_of_constantType : constantType -> string
*)