open Compiler_meta;;
open Instructions;;
open Instructions.JVM;;
open Instructions.DEX;;
(*open Java_insn;;
open Dex_insn;;*)
open Map_type;;
open Certificate;;

(*
let map_ifKind (kind: Java_insn.ifKind) =
	match kind with
		Java_insn.Eq -> Ifnz Dex_insn.Eq
	| Java_insn.Ne -> Ifnz Dex_insn.Ne
	| Java_insn.Gt -> Ifnz Dex_insn.Gt
	| Java_insn.Ge -> Ifnz Dex_insn.Ge
	| Java_insn.Lt -> Ifnz Dex_insn.Lt
	| Java_insn.Le -> Ifnz Dex_insn.Le
	| Java_insn.Eqz -> Ifz Dex_insn.Eq
	| Java_insn.Nez -> Ifz Dex_insn.Ne
	| Java_insn.Gtz -> Ifz Dex_insn.Gt
	| Java_insn.Gez -> Ifz Dex_insn.Ge
	| Java_insn.Ltz -> Ifz Dex_insn.Lt
	| Java_insn.Lez -> Ifz Dex_insn.Le
;;

let map_binopKind (kind: Java_insn.binop) =
	match kind with
		Java_insn.Add -> Dex_insn.Add
	| Java_insn.Sub -> Dex_insn.Sub
	| Java_insn.Mul -> Dex_insn.Mul
	| Java_insn.Div -> Dex_insn.Div
	| Java_insn.Mod -> Dex_insn.Mod
	| Java_insn.And -> Dex_insn.And
	| Java_insn.Or -> Dex_insn.Or
	| Java_insn.Xor -> Dex_insn.Xor
	| Java_insn.Shl -> Dex_insn.Shl
	| Java_insn.Shr -> Dex_insn.Shr
	| Java_insn.Ushr -> Dex_insn.Ushr
;;
*)

let min_register = 0;;

(* basically just make a register ordered from start with n number of      *)
(* elements                                                                *)
let rec makeRegList (start: int) (count: int) =
	match count <= 0 with
		true -> []
	| false -> start :: makeRegList (start + 1) (count - 1)
;;

let parseInsn (insn: JVM_Instruction.t) (m: dBlockMap) (bp: bpMap) (sb: sbMap) (ts: tsMap) 
    (ret: Block.t) (localN: int) (lvt: MethodPolicy.t) (constStackMap:valTypeMap) =
	let currentTS = Int32Map.find (JVM_Instruction.label32 insn) ts in
	let typeInfo = JVM_Instruction.typeInfo insn in
	let current_se = JVM_ins_typing.se typeInfo in
	let translated_rt = translate_st_rt (JVM_ins_typing.typing typeInfo) (localN) (lvt) (None) in
	let label32 = JVM_Instruction.label32 insn in
	let isStartBlock = Int32Map.find label32 sb in
	let blockIndex = Int32Map.find label32 bp in
	let cb = let cb_temp = Int32Map.find (blockIndex) m in
					match isStartBlock with
					| true -> Block.update_typing (current_se) (translated_rt) (cb_temp) 
					| false -> cb_temp in
	match JVM_Instruction.code insn with
	| JVM.Nop -> let size = match JVM_Instruction.size insn with | 1 -> 1 | 3 -> 2 
	                          | _ -> raise (Invalid_argument "[translate_java_program] invalid value of size for Nop")
													in  
		  (Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, 
	    DEX.createNopInsn (-1) (size) (current_se) (translated_rt)) cb) (m), min_register)
	| Pop -> (m, min_register)
	| Swap ->
			let rt1 = RT_op.add currentTS (RT_op.get (currentTS - 2) translated_rt) translated_rt in
			let rt2 = RT_op.add (currentTS + 1)	(RT_op.get (currentTS - 1) rt1) rt1 in
			let rt3 = RT_op.add (currentTS - 2)	(RT_op.get (currentTS + 1) rt2) rt2 in
			let newList = (*cb.insnList @ lastInsn @*) [
				(JVM_Instruction.label insn, DEX.createMoveInsn (-1) (1) (currentTS) (currentTS - 2) (current_se) (translated_rt));
				(JVM_Instruction.label insn, DEX.createMoveInsn (-1) (1) (currentTS + 1) (currentTS - 1) (current_se) (rt1));
				(JVM_Instruction.label insn, DEX.createMoveInsn (-1) (1) (currentTS - 2) (currentTS + 1) (current_se) (rt2));
				(JVM_Instruction.label insn, DEX.createMoveInsn (-1) (1) (currentTS - 1) (currentTS) (current_se) (rt3))]
			in
			(Int32Map.add (blockIndex) (Block.add_insns (newList) (cb)) (m), (currentTS + 1))
	| Dup ->
			(*(Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, DEX.createMoveInsn (-1) (1) (currentTS) (currentTS -1)
			  (current_se) (translated_rt)) cb) (m))*)
			let rt1 = RT_op.add (currentTS + 1) (RT_op.get (currentTS - 1) translated_rt) translated_rt in
			let rt2 = RT_op.add (currentTS - 1)	(RT_op.get (currentTS + 1) rt1) rt1 in
			let newList = (*cb.insnList @ lastInsn @*) [
				(JVM_Instruction.label insn, DEX.createMoveInsn (-1) (1) (currentTS + 1) (currentTS - 1) (current_se) (translated_rt));
				(JVM_Instruction.label insn, DEX.createMoveInsn (-1) (1) (currentTS - 1) (currentTS + 1) (current_se) (rt1));
				(JVM_Instruction.label insn, DEX.createMoveInsn (-1) (1) (currentTS - 2) (currentTS + 1) (current_se) (rt2))]
			in
			(Int32Map.add (blockIndex) (Block.add_insns (newList) (cb)) (m), (currentTS + 1))
	| JVM.Throw ->
			(Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, 
			DEX.createThrowInsn (-1) (1) (currentTS -1) (current_se) (translated_rt)) cb) (m), (currentTS - 1))
	| JVM.Push -> let arg = JVM.getArgsArgs insn in 
			let value = ArgsArgs.arg_at 0 arg in 
	     (Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, 
			 DEX.createConstInsn (-1) (1) (currentTS) (value) (current_se) (translated_rt)) cb) (m), currentTS)
	| JVM.Load -> let arg = JVM.getArgsArgs insn in
		  let value = ArgsArgs.arg_at 0 arg in 
		    (Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, 
				DEX.createMoveInsn (-1) (1) (currentTS) (value) (current_se) (translated_rt)) cb) (m), currentTS)
	| JVM.Store -> let arg = JVM.getArgsArgs insn in
		  let value = ArgsArgs.arg_at 0 arg in 
		    (Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, 
				DEX.createMoveInsn (-1) (1) (value) (currentTS - 1) (current_se) (translated_rt)) cb) (m), (currentTS-1))
	| JVM.IInc -> let arg = JVM.getArgsArgs insn in
		  let targetReg = ArgsArgs.arg_at 0 arg in 
		  let value = ArgsArgs.arg_at 1 arg in
				(Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, 
				DEX.createBinopConstInsn (-1) (1) (DEX.Add) (targetReg) (targetReg) (value) (current_se) 
				  (translated_rt)) cb) (m), targetReg)
	| JVM.Binop -> (let arg = JVM.getBinopArgs insn in
		  let binopKind = map_binopKind (JVM.BinopArgs.op arg) in
			let constStack = Int32Map.find label32 constStackMap in
			let ([c1; c2], _) = Utility.pop_n constStack 2 in
			match c1 || c2 with
			| false ->
				(Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, 
				DEX.createBinopInsn (-1) (1) (binopKind) (currentTS-2) (currentTS-2) (currentTS-1)
				(current_se) (translated_rt)) cb) (m), (currentTS-1))
			| true ->
				let (r, c) = match c2 with
				| true -> (currentTS-1, currentTS-2) (* at the moment we don't care of the constant value *)
				| false -> (currentTS-2, currentTS-1) in
				(Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, 
				DEX.createBinopConstInsn (-1) (1) (binopKind) (currentTS-2) (r) (c)
				(current_se) (translated_rt)) cb) (m), (currentTS-1))) (* TODO: recheck whether this is the correct maximum register *)
	| JVM.If -> (let arg = JVM.getIfArgs insn in
		  let target = ((JVM_Instruction.label insn) + (JVM.IfArgs.offset arg)) in
	    let jvmIfKind = JVM.IfArgs.kind arg in
		  let dexIfKind = map_ifKind (jvmIfKind) in
			match jvmIfKind with
				| Eq | Ne | Gt | Ge | Lt | Le -> (Int32Map.add (blockIndex)
								(Block.add_insn (JVM_Instruction.label insn, DEX.createIfInsn (-1) (2) (dexIfKind) 
								(currentTS - 2) (currentTS - 1) (target) (current_se) (translated_rt)) cb) m, (currentTS-1))
				| Eqz | Nez | Gtz | Gez | Ltz | Lez -> (Int32Map.add (blockIndex)
				        (Block.add_insn (JVM_Instruction.label insn, DEX.createIfzInsn (-1) (2) (dexIfKind) 
								(currentTS - 1) (target) (current_se) (translated_rt)) cb) (m), (currentTS-1)) 
				)						
	| JVM.Goto -> (match isStartBlock with 
	    | true -> (Int32Map.add (blockIndex) (cb) (m), min_register)
			| false -> (m, min_register)) 
	| JVM.Return -> (let arg = JVM.getReturnArgs insn in 
		  match JVM.ReturnArgs.retType arg with
					JVM.Void -> (m, min_register)
				| JVM.Int -> (Int32Map.add (blockIndex) (Block.add_insn 
					  (JVM_Instruction.label insn, DEX.createMoveInsn (-1) (1) (0) 
						(currentTS - 1) (current_se) (RT_op.update_flag (1) (translated_rt))) cb) (m), (currentTS-1))
			)
	| JVM.New -> let arg = JVM.getClassArgs insn in
		      (Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, 
					DEX.createNewInsn (-1) (1) (currentTS) (JVM.ClassArgs.name arg) (current_se) 
					(translated_rt)) cb) m, currentTS)
	| JVM.Getfield -> let arg = JVM.getFMIArgs insn in
	            let fieldName = (JVM.FMIArgs.class_name arg) ^ "." ^ (JVM.FMIArgs.name arg) in
		          (Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, 
		          DEX.createGetfieldInsn (-1) (1) (currentTS -1) (currentTS - 1) (fieldName) (current_se) (translated_rt)) cb) m, (currentTS-1))
  | JVM.Putfield -> let arg = JVM.getFMIArgs insn in
	            let fieldName = (JVM.FMIArgs.class_name arg) ^ "." ^ (JVM.FMIArgs.name arg) in
		          (Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, 
		          DEX.createPutfieldInsn (-1) (1) (currentTS - 1) (currentTS - 2) (fieldName) (current_se) (translated_rt)) cb) m, (currentTS-1))
	| JVM.Invoke t -> let arg = JVM.getFMIArgs insn in
	        let methodName = (JVM.FMIArgs.class_name arg) ^ "." ^ (JVM.FMIArgs.name arg) in
		      let nArgs = (let close = String.index (JVM.FMIArgs.fmi_type arg) (')') in
							let str = String.sub (JVM.FMIArgs.fmi_type arg) (1) (close - 1) in
							match (String.length (str)) = 0 with
								false -> (Utility.countArgs (str) (String.length (str)) (0))
							| true -> (0)) in
					let regList = makeRegList (currentTS - nArgs - 1) (nArgs + 1) in
						(Int32Map.add (blockIndex) (Block.add_insn (JVM_Instruction.label insn, 
						DEX.createInvokeInsn (-1) (1) (nArgs) (regList) (methodName) (current_se) 
						(translated_rt)) cb) (m), (currentTS-1))
	(*| _ -> raise (Invalid_argument "Bogus code for Plain Instruction")*)
(*
	match insn with
		Plain_insn ins ->
			( let translated_rt = translate_st_rt (ins.jvm_typeInfo.st) (localN) (lvt) in
				let se = ins.jvm_typeInfo.jvm_se in
				let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
				let blockIndex = Int32Map.find (Int32.of_int ins.label) (bp) in
				let isStartBlock = Int32Map.find (Int32.of_int ins.label) (sb) in
				let cb = let cb_temp = Int32Map.find (blockIndex) m in
					match isStartBlock with
					| true -> { cb_temp with block_se = se; rt0 = translated_rt }
					| false -> cb_temp in
				let lastInsn = match cb.lastInsn with
						None -> [] | Some ins -> [ins]
				in
				(match ins.code with
						Pop -> (m)
					| Swap ->
							let rt1 = Int32Map.add (Int32.of_int currentTS)
									(Int32Map.find (Int32.of_int (currentTS - 2)) translated_rt) translated_rt in
							let rt2 = Int32Map.add (Int32.of_int (currentTS + 1))
									(Int32Map.find (Int32.of_int (currentTS - 1)) translated_rt) rt1 in
							let rt3 = Int32Map.add (Int32.of_int (currentTS - 2))
									(Int32Map.find (Int32.of_int (currentTS + 1)) translated_rt) rt2 in
							let newList = cb.insnList @ lastInsn @ [
								create_MoveDexInsn (-1) (1) (currentTS, (currentTS - 2)) ({ dex_se = se; rt = translated_rt });
								create_MoveDexInsn (-1) (1) ((currentTS + 1), (currentTS - 1)) ({ dex_se = se; rt = rt1 });
								create_MoveDexInsn (-1) (1) ((currentTS - 2), (currentTS + 1)) ({ dex_se = se; rt = rt2 })]
							in
							let newLast = create_MoveDexInsn (-1) (1) ((currentTS - 1), currentTS) ({ dex_se = se; rt = rt3 }) in
							(Int32Map.add (blockIndex) { cb with insnList = newList; lastInsn = Some newLast } (m))
					| Dup ->
							(Int32Map.add (blockIndex) { cb with insnList = (cb.insnList @ lastInsn);
										lastInsn = Some (create_MoveDexInsn (-1) (1) ((currentTS), currentTS - 1)
													({ dex_se = se; rt = translated_rt })) } (m))
					| Java_insn.Throw ->
							(Int32Map.add (blockIndex) { cb with insnList = (cb.insnList @ lastInsn);
										lastInsn = Some (create_ArgDexInsn (-1) (1) (Throw) (currentTS - 1)
													({ dex_se = se; rt = translated_rt })) } (m))
					| _ -> raise (Invalid_argument "Bogus code for Plain Instruction"))
			)
	| Args_insn ins ->
			( let translated_rt = translate_st_rt (ins.jvm_typeInfo.st) (localN) (lvt) in
				let se = ins.jvm_typeInfo.jvm_se in
				let typeInfo = { dex_se = se; rt = translated_rt } in
				let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
				let blockIndex = Int32Map.find (Int32.of_int ins.label) (bp) in
				let isStartBlock = Int32Map.find (Int32.of_int ins.label) (sb) in
				let cb = let cb_temp = Int32Map.find (blockIndex) m in
					match isStartBlock with
					| true -> { cb_temp with block_se = se; rt0 = translated_rt }
					| false -> cb_temp in
				let lastInsn = match cb.lastInsn with None -> [] | Some ins -> [ins] in
				(match ins.code with
						Push -> (Int32Map.add (blockIndex)
									{ cb with insnList = (cb.insnList @ lastInsn); lastInsn =
											Some (create_ConstDexInsn (-1) (1) (currentTS , (List.hd ins.args)) (typeInfo)) } (m))
					| Load -> (Int32Map.add (blockIndex)
									{ cb with insnList = (cb.insnList @ lastInsn); lastInsn =
											Some (create_MoveDexInsn (-1) (1) (currentTS, (List.hd ins.args)) (typeInfo)) } (m))
					| Store -> (Int32Map.add (blockIndex)
									{ cb with insnList = (cb.insnList @ lastInsn); lastInsn =
											Some (create_MoveDexInsn (-1) (1) ((List.hd ins.args), (currentTS - 1)) (typeInfo)) } (m))
					| IInc ->
							let targetReg = List.hd ins.args in
							let value = List.hd (List.tl ins.args) in
							(Int32Map.add (blockIndex)
									{ cb with insnList = (cb.insnList @ lastInsn); lastInsn =
											Some (create_BinopDexInsn (-1) (1) ((Dex_insn.Add), (targetReg), (targetReg), (value), true) (typeInfo)) } (m))
					| _ -> raise (Invalid_argument "Bogus code for Args Instruction"))
			)
	| Binop_insn ins ->
			( let translated_rt = translate_st_rt (ins.jvm_typeInfo.st) (localN) (lvt) in
				let se = ins.jvm_typeInfo.jvm_se in
				let typeInfo = { dex_se = se; rt = translated_rt } in
				let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
				let blockIndex = Int32Map.find (Int32.of_int ins.label) (bp) in
				let cb = Int32Map.find (blockIndex) m in
				let lastInsn = match cb.lastInsn with None -> [] | Some ins -> [ins] in
				let binopKind = map_binopKind (ins.args.op) in
				Int32Map.add (blockIndex)
					{ cb with insnList = (cb.insnList @ lastInsn); lastInsn =
							Some (create_BinopDexInsn (-1) (1) (binopKind, (currentTS - 2), (currentTS - 2),
										(currentTS - 1), false) (typeInfo)) } (m)
			)
	| If_insn ins ->
			( let translated_rt = translate_st_rt (ins.jvm_typeInfo.st) (localN) (lvt) in
				let se = ins.jvm_typeInfo.jvm_se in
				let typeInfo = { dex_se = se; rt = translated_rt } in
				let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
				let blockIndex = Int32Map.find (Int32.of_int ins.label) (bp) in
				let isStartBlock = Int32Map.find (Int32.of_int ins.label) (sb) in
				let cb = let cb_temp = Int32Map.find (blockIndex) m in
					match isStartBlock with
					| true -> { cb_temp with block_se = se; rt0 = translated_rt }
					| false -> cb_temp in
				let target = ins.label + ins.args.targetOffset in
				let lastInsn = match cb.lastInsn with None -> [] | Some ins -> [ins] in
				let dexIfKind = map_ifKind (ins.args.kind) in
				match dexIfKind with
					Ifnz _ -> (Int32Map.add (blockIndex)
								{ cb with insnList = (cb.insnList @ lastInsn); lastInsn =
										Some (create_IfDexInsn (-1) (2) (dexIfKind) ((currentTS - 2), (currentTS - 1), (target)) (typeInfo)) } (m))
				| Ifz _ -> (Int32Map.add (blockIndex)
								{ cb with insnList = (cb.insnList @ lastInsn); lastInsn =
										Some (create_IfzDexInsn (-1) (2) (dexIfKind) ((currentTS - 1), (target)) (typeInfo)) } (m))
			)
	(* The goto instruction will be removed, and will be added later for all *)
	(* the block to maintain the invariant that each block will end in sort  *)
	(* of branching instruction                                              *)
	| Goto_insn ins ->
			( let isStartBlock = Int32Map.find (Int32.of_int ins.label) (sb) in
				match isStartBlock with
				| true -> (
							let translated_rt = translate_st_rt (ins.jvm_typeInfo.st) (localN) (lvt) in
							let se = ins.jvm_typeInfo.jvm_se in
							let cb = Int32Map.find (Int32.of_int ins.label) (m) in
							Int32Map.add (Int32.of_int ins.label)
								{ cb with block_se = se; rt0 = translated_rt } (m) )
				| false -> (m)
			)
	| Return_insn ins ->
			( match ins.args.retType with
					Java_insn.Void -> (m)
				| _ -> (
							let translated_rt = translate_st_rt (ins.jvm_typeInfo.st) (localN) (lvt) in
							let se = ins.jvm_typeInfo.jvm_se in
							let typeInfo = { dex_se = se; rt = translated_rt } in
							let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
							let blockIndex = Int32Map.find (Int32.of_int ins.label) (bp) in
							let isStartBlock = Int32Map.find (Int32.of_int ins.label) (sb) in
							let cb = let cb_temp = Int32Map.find (blockIndex) m in
								match isStartBlock with
								| true -> { cb_temp with block_se = se; rt0 = translated_rt }
								| false -> cb_temp in
							let lastInsn = match cb.lastInsn with None -> [] | Some ins -> [ins] in
							(Int32Map.add (blockIndex) { cb with insnList = (cb.insnList @ lastInsn); lastInsn = Some
											(create_MoveDexInsn (-1) (1) (0 , (currentTS - 1)) (typeInfo)) } (m))
						)
			)
	| New_insn ins ->
			let translated_rt = translate_st_rt (ins.jvm_typeInfo.st) (localN) (lvt) in
			let se = ins.jvm_typeInfo.jvm_se in
			let typeInfo = { dex_se = se; rt = translated_rt } in
			let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
			let blockIndex = Int32Map.find (Int32.of_int ins.label) (bp) in
			let isStartBlock = Int32Map.find (Int32.of_int ins.label) (sb) in
			let cb = let cb_temp = Int32Map.find (blockIndex) m in
				match isStartBlock with
				| true -> { cb_temp with block_se = se; rt0 = translated_rt }
				| false -> cb_temp in
			let lastInsn = match cb.lastInsn with None -> [] | Some ins -> [ins] in
			(Int32Map.add (blockIndex)
					{ cb with insnList = (cb.insnList @ lastInsn); lastInsn =
							Some (create_NewDexInsn (-1) (1) (currentTS , (ins.args.class_name)) (typeInfo)) } (m))
	| FMI_insn ins ->
			let translated_rt = translate_st_rt (ins.jvm_typeInfo.st) (localN) (lvt) in
			let se = ins.jvm_typeInfo.jvm_se in
			let typeInfo = { dex_se = se; rt = translated_rt } in
			let currentTS = Int32Map.find (Int32.of_int ins.label) (ts) in
			let blockIndex = Int32Map.find (Int32.of_int ins.label) (bp) in
			let isStartBlock = Int32Map.find (Int32.of_int ins.label) (sb) in
			let cb = let cb_temp = Int32Map.find (blockIndex) m in
				match isStartBlock with
				| true -> { cb_temp with block_se = se; rt0 = translated_rt }
				| false -> cb_temp in
			let lastInsn = match cb.lastInsn with None -> [] | Some ins -> [ins] in
			match ins.code with
				Putfield -> (Int32Map.add (blockIndex)
							{ cb with insnList = (cb.insnList @ lastInsn); lastInsn =
									Some (create_FieldDexInsn (-1) (1) (Field Put) ((currentTS - 1), (currentTS - 2), (ins.args.fmi_class_name ^ "." ^ ins.args.fmi_name))
											(typeInfo)) } (m))
			| Getfield -> (Int32Map.add (blockIndex)
							{ cb with insnList = (cb.insnList @ lastInsn); lastInsn =
									Some (create_FieldDexInsn (-1) (1) (Field Get) ((currentTS - 1), (currentTS - 1), (ins.args.fmi_class_name ^ "." ^ ins.args.fmi_name))
											(typeInfo)) } (m))
			| Java_insn.Invoke t ->
					let nArgs = (let close = String.index (ins.args.fmi_type) (')') in
							let str = String.sub (ins.args.fmi_type) (1) (close - 1) in
							match (String.length (str)) = 0 with
								false -> (countArgs (str) (String.length (str)) (0))
							| true -> (0)) in
					let regList = makeRegList (currentTS - nArgs - 1) (nArgs + 1) ([]) in
					(match t with
							Java_insn.Int ->
								(Int32Map.add (blockIndex)
										{ cb with insnList = (cb.insnList @ lastInsn);
											lastInsn = Some (create_MethodDexInsn (-1) (1) ((nArgs), (regList),
															(ins.args.fmi_class_name ^ "." ^ ins.args.fmi_name)) (typeInfo)) } (m))
						| Java_insn.Void ->
								(Int32Map.add (blockIndex)
										{ cb with insnList = (cb.insnList @ lastInsn);
											lastInsn = Some (create_MethodDexInsn (-1) (1) ((nArgs), (regList),
															(ins.args.fmi_class_name ^ "." ^ ins.args.fmi_name)) (typeInfo)) } (m))
					)
			| _ -> raise (Invalid_argument ("Not a proper code for field manipulation / method instruction : " ^ string_of_insnType(ins.code)))
;; (*| _ -> raise (Invalid_argument "Under development");;*)
*)

let rec parseInsns (jProg: JVM_Instruction.t list) (m: dBlockMap) (bp: bpMap) (sb: sbMap)
		(ts: tsMap) (ret: Block.t) (localN: int) (lvt: MethodPolicy.t) (constStackMap: valTypeMap) =
	match jProg with
		[] -> (m, min_register)
	| insn :: t ->
			let (newM, max_registers) = parseInsn (insn) (m) (bp) (sb) (ts) (ret) (localN) (lvt) (constStackMap) in
			let (newM, max_registers') = parseInsns (t) (newM) (bp) (sb) (ts) (ret) (localN) (lvt) (constStackMap) in
			(newM, match max_registers > max_registers' with | true -> max_registers | false -> max_registers');;

let parse (jProg: JVM_Instruction.t list) (m: dBlockMap) (bp: bpMap) (sb : sbMap) (ts: tsMap) 
  (ret: Block.t) (localN: int) (lvt: MethodPolicy.t) (constStackMap: valTypeMap) =
	parseInsns (jProg) (m) (bp) (sb) (ts) (ret) (localN) (lvt) (constStackMap) ;;